package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.config.properties.MaterialTypeProperties
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.model.validation.isNotNullAndNotBlank
import dev.fyloz.colorrecipesexplorer.repository.MaterialTypeRepository
import org.springframework.context.annotation.Profile
import org.springframework.stereotype.Service

interface MaterialTypeService :
  ExternalNamedModelService<MaterialType, MaterialTypeSaveDto, MaterialTypeUpdateDto, MaterialType, MaterialTypeRepository> {
    /** Checks if a material type with the given [prefix] exists. */
    fun existsByPrefix(prefix: String): Boolean

    /** Checks if the given [materialType] is used by one or more materials. */
    fun isUsedByMaterial(materialType: MaterialType): Boolean

    /** Gets all system material types. */
    fun getAllSystemTypes(): Collection<MaterialType>

    /** Gets all material types who are not a system type. */
    fun getAllNonSystemType(): Collection<MaterialType>

    /** Saves and update the system material types. */
    fun saveSystemTypes(systemTypeProperties: Collection<MaterialTypeProperties.MaterialTypeProperty>)
}

@Service
@Profile("!emergency")
class MaterialTypeServiceImpl(repository: MaterialTypeRepository, private val materialService: MaterialService) :
  AbstractExternalNamedModelService<MaterialType, MaterialTypeSaveDto, MaterialTypeUpdateDto, MaterialType, MaterialTypeRepository>(
    repository
  ), MaterialTypeService {
    override fun idNotFoundException(id: Long) = materialTypeIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = materialIdAlreadyExistsException(id)
    override fun nameNotFoundException(name: String) = materialTypeNameNotFoundException(name)
    override fun nameAlreadyExistsException(name: String) = materialTypeNameAlreadyExistsException(name)

    override fun MaterialType.toOutput() = this

    override fun existsByPrefix(prefix: String): Boolean = repository.existsByPrefix(prefix)
    override fun isUsedByMaterial(materialType: MaterialType): Boolean =
      materialService.existsByMaterialType(materialType)

    override fun getAllSystemTypes(): Collection<MaterialType> = repository.findAllBySystemTypeIs(true)
    override fun getAllNonSystemType(): Collection<MaterialType> = repository.findAllBySystemTypeIs(false)

    override fun save(entity: MaterialType): MaterialType {
        if (existsByPrefix(entity.prefix))
            throw materialTypePrefixAlreadyExistsException(entity.prefix)
        return super<AbstractExternalNamedModelService>.save(entity)
    }

    override fun update(entity: MaterialTypeUpdateDto): MaterialType {
        val persistedMaterialType by lazy { getById(entity.id) }

        return update(with(entity) {
            MaterialType(
              id = id,
              name = if (isNotNullAndNotBlank(name)) name else persistedMaterialType.name,
              prefix = if (isNotNullAndNotBlank(prefix)) prefix else persistedMaterialType.prefix,
              systemType = false
            )
        })
    }

    override fun update(entity: MaterialType): MaterialType {
        with(repository.findByPrefix(entity.prefix)) {
            if (this != null && id != entity.id)
                throw materialTypePrefixAlreadyExistsException(entity.prefix)
        }

        return super<AbstractExternalNamedModelService>.update(entity)
    }

    override fun delete(entity: MaterialType) {
        if (!repository.canBeDeleted(entity.id!!)) throw cannotDeleteMaterialTypeException(entity)
        super.delete(entity)
    }

    override fun saveSystemTypes(systemTypeProperties: Collection<MaterialTypeProperties.MaterialTypeProperty>) {
        val systemTypes = systemTypeProperties.map { it.toMaterialType() }
        val oldSystemTypes = getAllSystemTypes().toMutableSet()

        fun saveOrUpdateSystemType(type: MaterialType) {
            if (existsByName(type.name)) {
                val persistedMaterialType = getByName(type.name)
                update(materialType(type, newId = persistedMaterialType.id, newSystemType = true))
            } else {
                save(type)
            }
        }

        // Save new system types
        systemTypes.forEach {
            saveOrUpdateSystemType(it)
            oldSystemTypes.removeIf { c -> c.name == it.name }
        }

        // Remove old system types
        oldSystemTypes.forEach { update(materialType(it, newSystemType = false)) }
    }
}
